@@ -4,15 +4,20 @@ import requests
 from enum import Enum
 from functools import lru_cache
 from requests.exceptions import ReadTimeout, Timeout
+import os
+from .exceptions import NotAuthorised, MissingClientId, MissingClientSecret, MissingTicker
 
 FF_LOGIN_URI = "oauth/login"
 
+os.environ['NO_PROXY'] = 'localhost'
+
 """
 Financefeast client API library
 https://financefeast.io
 """
 
 class Environments(Enum):
+    dev = "http://localhost:5001"
     test = "https://api.test.financefeast.io"
     prod = "https://api.financefeast.io"
 
@@ -26,7 +31,6 @@ class FinanceFeast:
         self._token = token
         self._logger = logger
         self._environment = environment
-        self._access_token = None
 
         if not logger:
             self._logger = logging.getLogger('ff_client')
@@ -44,12 +48,12 @@ class FinanceFeast:
             self._client_secret = os.environ.get('FF-CLIENT-SECRET')
 
         if not self._client_secret and not self._token:
-            raise Exception(
+            raise MissingClientSecret(
                 "parameter 'client_id' must be either passed or set as an environment variable 'FF-CLIENT-ID', or pass parameter 'token' with a valid bearer token"
             )
 
         if not self._client_id and not self._token:
-            raise Exception(
+            raise MissingClientId(
                 "parameter 'client_secret' must be either passed or set as an environment variable 'FF-CLIENT-SECRET', or pass parameter 'token' with a valid bearer token"
             )
 
@@ -57,15 +61,18 @@ class FinanceFeast:
             self._logger.debug(f'Authorizing to Financefeast API environment {self._environment}')
             self.__authorize()
         else:
-            self._logger.debug(f'Authorized using supplied token to Financefeast API environment {self._environment}')
-            self._access_token = self._token
+            validated = self._check_authorization()
+            if not validated:
+                self.__authorize()
+            else:
+                self._logger.debug(f'Authorized using supplied token to Financefeast API environment {self._environment}')
 
     def __authorize(self):
         """
         Authorize client credentials
         :return: access token
         """
-        if not self._access_token:
+        if not self._token and self._client_id and self._client_secret:
             url = f'{self._environment.value}/{FF_LOGIN_URI}'
             self._logger.debug(f'Constructed url {url} for authorization')
 
@@ -74,7 +81,7 @@ class FinanceFeast:
             payload = self._requests.get(url=url, headers=headers)
 
             try:
-                self._access_token = payload['access_token']
+                self._token = payload['access_token']
                 self._logger.debug('Found a valid access_token')
             except KeyError:
                 self._logger.exception(f'{payload["detail"]}')
@@ -82,12 +89,32 @@ class FinanceFeast:
                     f'{payload["detail"]}'
                 )
 
-            self._logger.info("Client successfully authorized to API")
+            self._logger.info("Client successfully authorized to API using client credentials")
+            return self._token
+
+        self._logger.warning("No client_id, client_secret or an invalid token has been submitted. Pass a valid token or supply your client credentails to authorize to the Financefeast API")
+        raise NotAuthorised()
+
+
+    def _check_authorization(self):
+        """
+        Check a token is valid by calling the validate endpoint
+        :return:
+        """
+        self._logger.debug("Introspecting token")
+        introspected_token_result = self.validate()
+
+        if not introspected_token_result:
+            self._token = None
+            self._logger.warning(f"Token is not valid or has expired.")
+            return False
+
+        self._logger.debug("Token is valid")
+        return True
 
-        return self._access_token
 
     def _generate_authorization_header(self):
-        return {'Authorization': f'Bearer {self._access_token}'}
+        return {'Authorization': f'Bearer {self._token}'}
 
 
     """
@@ -227,7 +254,7 @@ class FinanceFeast:
 
         # check required parameters
         if not ticker:
-            raise Exception(
+            raise MissingTicker(
                 "parameter `ticker` must be either passed"
             )
 
@@ -270,7 +297,7 @@ class FinanceFeast:
 
         # check required parameters
         if not ticker:
-            raise Exception(
+            raise MissingTicker(
                 "parameter `ticker` must be either passed"
             )
 
@@ -314,7 +341,7 @@ class FinanceFeast:
 
         # check required parameters
         if not ticker:
-            raise Exception(
+            raise MissingTicker(
                 "parameter `ticker` must be either passed"
             )
 
@@ -361,7 +388,7 @@ class FinanceFeast:
 
         # check required parameters
         if not ticker:
-            raise Exception(
+            raise MissingTicker(
                 "parameter `ticker` must be either passed"
             )
 
@@ -407,7 +434,7 @@ class FinanceFeast:
 
         # check required parameters
         if not ticker:
-            raise Exception(
+            raise MissingTicker(
                 "parameter `ticker` must be either passed"
             )
 
@@ -451,7 +478,7 @@ class FinanceFeast:
 
         # check required parameters
         if not ticker:
-            raise Exception(
+            raise MissingTicker(
                 "parameter `ticker` must be either passed"
             )
 
@@ -499,7 +526,7 @@ class FinanceFeast:
 
         # check required parameters
         if not ticker:
-            raise Exception(
+            raise MissingTicker(
                 "parameter `ticker` must be either passed"
             )
 
@@ -549,7 +576,7 @@ class FinanceFeast:
 
         # check required parameters
         if not ticker:
-            raise Exception(
+            raise MissingTicker(
                 "parameter `ticker` must be either passed"
             )
 
@@ -597,7 +624,7 @@ class FinanceFeast:
 
         # check required parameters
         if not ticker:
-            raise Exception(
+            raise MissingTicker(
                 "parameter `ticker` must be either passed"
             )
 
@@ -633,7 +660,7 @@ class FinanceFeast:
 
     class RequestRateLimited():
         TIMEOUT_CONN = 1.5
-        TIMEOUT_RESP = 1.5
+        TIMEOUT_RESP = 5
         RATE_LIMIT_HEADER_LIMIT_NAME = 'x-ratelimit-limit'
         RATE_LIMIT_HEADER_REMAINING_NAME = 'x-ratelimit-remaining'
         RATE_LIMIT_HEADER_RESET_NAME = 'x-ratelimit-reset'
@@ -677,9 +704,7 @@ class FinanceFeast:
             try:
                 r = requests.get(*args, timeout=(self.TIMEOUT_CONN, self.TIMEOUT_RESP), **kwargs)
             except (ReadTimeout, Timeout) as e:
-                raise Exception(
-                    f"Timeout triggered calling {kwargs.get('url')}"
-                )
+                raise NotAuthorised()
 
             # inspect rate limits
             self.logger.debug(f'Parsing rate limit headers')
@@ -689,9 +714,7 @@ class FinanceFeast:
                 f'Request to {kwargs.get("url")} returned a {r.status_code} status code')
 
             if r.status_code == 403:
-                raise Exception(
-                    "Not authorized"
-                )
+                raise NotAuthorised()
 
             if r.status_code == 429:
                 self.logger.error(f'Rate limited exceeded. {r.json()}')
